/* UsiFe FileSystem
----------------
Copyright (2012): Rohan Sharma and Saurabh Kumar

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

CONTACT: 
Rohan Sharma : rickyrohan14@gmail.com
Saurabh Kumar : saurabh.catch@gmail.com
Smruti R. Sarangi : srsarangi@cse.iitd.ac.in
*/
#include "fuse_include.h" 
#include "do_write.h"

/*
 *Truncate file to a given length.
 *Data stored about encrypted regions need to be changed according to the changed
 *file size.
 *This function retains data only for the blocks 
 *remaining in the file(excluding the last data block if it has been cut in between,
 *which should be corrected).
 *This function uses linear search right now(just for simplicity).
 */

/*
 *"i_file_acl" field of inode stores link to a block which stores data about encrypted regions.
 * 
 */

/* STORED DATA about encrypted regions.
 * For every region we store following structure on a disk block. Each block should be of size 4096 bytes.
 *  struct region1{
 *          int block; //block no of the encrypted region
 *          short start;//starting offset within the block.
 *          short end; //ending offset within the block.
 *         };  	
 *  Since size of struct is 8 bytes  , so we can have 4096/8=512 encrypted regions in a single block.	
 *  Also first two bytes of block store number of encrypted regions , so we can have maximum of 511 encrypted regions.
 *  But since number of encrypted regions can be more then 511 ,we don't store 511 regions in a single block, rather 
 *  in case no. of regions exceed 510 a new block is allocated and last 4 bytes of the current block then 
 *  link to the new block.Remaining regions are then stored in next block; 
 * 
 *  First two bytes of the block will then have value 511(though we are storing only 510 entries). 
 *  
 *  Also to speed up things we store no of bytes encrypted by all entries in the current block at offset 4088, just before
 *  link to the next block.
 *
 *
 *  OVERALL LAYOUT (4096 size block)
 *
 *  [2 bytes for no. of entries(It has value 511 in case it links to next block)]
 *  [Encrypted region entries...]
 *  [Encrypted region entries...]
 *  .............................   
 *  .............................
 *  .............................
 *  .............................
 *  [Number of bytes encrypted at offset 4088]
 *  [Link to next block in case one exists.] 
 *
 *
 *
 */
int do_truncate(const char *path, off_t length)
{
	int rt;
	ext2_file_t efile;
 	ext2_filsys e2fs = (ext2_filsys)current_ext2fs();
        ext2_ino_t ino;
	struct ext2_inode inode;

	DEBUG_PRINTF("Truncating \n %s  %d ",path,(int)length); 

        rt = do_readinode(e2fs, path, &ino, &inode);
	if (rt) {
		DEBUG_PRINTF("do_readinode %s failed\n", path);
		return rt;
	}  

        if(inode.i_size==length)
         {
            DEBUG_PRINTF("No need to truncate\n");
            return 0;
         }
       
	efile = do_open2((ext2_filsys)e2fs, path, O_WRONLY);
	if (efile == NULL) {
		DEBUG_PRINTF("Failed to open file %s\n", path);
		return -ENOENT;
	}
	
	/*
	 *Change file size , it should have been done later.
	 * 
 	 */
        DEBUG_PRINTF("SET SIZE CALLING \n"); 
	DEBUG_PRINTF("before size %d\n",inode.i_size); 
        DEBUG_PRINTF("Inode %d %d\n",efile->inode,inode);
        rt = ext2fs_file_set_size(efile, length);
        ext2fs_file_flush(efile);
        DEBUG_PRINTF("after size %d %d\n",ext2fs_file_get_size(efile),rt); 
        
        rt = do_readinode(e2fs, path, &ino, &inode);
        printf("after size %d \n",inode.i_size); 

	/*
	 * i_file_acl==0 implies no encrypted regions.
	 */ 
         if(inode.i_file_acl!=0)
         {
           
          char data[4096];//block data
          memset(data,0,4096);
          int block=length/4096;
          short index=0;//Number of entries
          int *p; 
           
          if(length==0) //truncate to length zero
            {
                int acl_block;//link to next block
                io_channel_read_blk(e2fs->io,inode.i_file_acl,1, data);  //function for reading directly from physical block on disk
				    
                memcpy(&index,data,sizeof(short));
                memcpy(&acl_block,data+4092,sizeof(int)); 
             
                memset(data,0,4096);
                io_channel_write_blk(e2fs->io,inode.i_file_acl, 1,data);
                DEBUG_PRINTF("inode.i_file_acl truncated %d %d %d\n",acl_block,index,e2fs->block_map); 
                
		while(index==511){ //index==511 means there is a next block
                      
		      int previous=acl_block;
                      DEBUG_PRINTF("Truncating next block %d",acl_block);
                      io_channel_read_blk(e2fs->io,acl_block,  //function for reading directly from physical block on disk
				    1, data);
                      memcpy(&index,data,sizeof(short));
                      memcpy(&acl_block,data+4092,sizeof(int)); 
                      ext2fs_block_alloc_stats(e2fs,previous, -1); //deallocate block 
                    }

	       /*
	 	* Tests if deallocations worked correctly.Only for debugging.
	 	*/
	       
  	       /* 			
                *rt=ext2fs_test_block_bitmap(e2fs->block_map, 4812);
                *DEBUG_PRINTF("TEST %d",rt);
                *rt=ext2fs_test_block_bitmap(e2fs->block_map, 4809);
                *DEBUG_PRINTF("TEST %d",rt);
                */
	
               rt=ext2fs_write_block_bitmap(e2fs);
               if(rt)
                 DEBUG_PRINTF("BITMAP WRITE FAILED\n\n");
 
                 }
	       
		else
		{               
                io_channel_read_blk(e2fs->io,inode.i_file_acl,1, data);  //function for reading directly from physical block on disk
                blk_t acl_block=inode.i_file_acl;
                int *p,k;
                short items=0; 

                memcpy(&items,data,sizeof(short)); 
                
		for(index=0;index<items;)                                       //using linear search right 
                {
                 DEBUG_PRINTF("Truncate %d %d\n\n",index,items);
                 p=(int *)(data+2+index*8);
                 if((*p)>block)
                   break;

                 index++;
                 if(index==510 && items==511)
                      {
                            DEBUG_PRINTF("LOADING NEW ACL BLOCK %d\n\n",acl_block);
                            p=(int *)(data+4092); 
                            io_channel_read_blk(e2fs->io,*p,1, data); 
                            acl_block=(blk_t)*p;
                            DEBUG_PRINTF("New ACl BLOCK %d\n\n",acl_block);  
                            index=0;
                            memcpy(&items,data,sizeof(short));   
                      }
                  }                                                
 
             
            	 memcpy(data,&index,sizeof(short)); 
               	 io_channel_write_blk(e2fs->io,acl_block, 1,data);

               /*
                *Freeing up unused blocks
                */  
           
           	while(items==511){
                            DEBUG_PRINTF("Freeing blocks\n\n");
                            p=(int *)(data+4092); 
                            io_channel_read_blk(e2fs->io,*p,1, data); 
				    
                            memcpy(&items,data,sizeof(short));
                            ext2fs_block_alloc_stats(e2fs, *p, -1);  
                               }
           	}
                DEBUG_PRINTF("index truncated to %d\n\n",index); 
          }

        if (rt) {
		
		DEBUG_PRINTF("Truncation failed %s\n ",path);
		return rt;
	}

	rt = do_release2(efile);
	if (rt != 0) {
		DEBUG_PRINTF("do_release(efile); failed\n");
		return rt;
	}

	fflush(stdout);
	return 0;
}

int do_write(const char *path, const char *buf, size_t size, off_t offset, struct fuse_file_info *fi)
{
        int encryt_start =  10;
        int encryt_end   =  521;
	int rt;
	const char *tmp;
	unsigned int wr,wr1=0;
	unsigned long long npos;
	int fsize;

	ext2_ino_t ino;
	struct ext2_inode inode;
	ext2_filsys e2fs = (ext2_filsys) current_ext2fs();
        ext2_file_t efile; 
 
	DEBUG_PRINTF("Writing to file %s %d %d\n",path,(int)offset,(int)size);

	if(!gtk_override)
	   {
	    PWord_t  PValue; 
	    int val;
	    ext2_file_t efile = (ext2_file_t)((unsigned long)(fi->fh));
	    JLG(PValue,PJLArray,efile->ino);
	    if(PValue!=PJERR && PValue!=NULL)
	      {
		DEBUG_PRINTF("PValue is %d\n",*PValue);
		val=((*PValue)&ENC_MODE);
		DEBUG_PRINTF("Enc mode is %d\n",val);
		e=val;
	      }
            else if(PValue==NULL)	
	       {
		DEBUG_PRINTF("PValue is null\n");
		e=0;
      		}
    	     else
	      {
		DEBUG_PRINTF("Hashing error\n");
        	return -1;
	      }		
		
	   }
        else
         {
        /*
	 *Only for debugging.
	 */	
	if(e1==1)
          {  
          if(((rand()*1.0)/(double)RAND_MAX)<e_p)             
             e=1;
          else
             e=0;
          }
        else if(e1==0)
          {
           e=0;
          } 
         }
        efile = (ext2_file_t)((unsigned long)(fi->fh));
        
	DEBUG_PRINTF("efile %d\n\n",(int)efile);

        pthread_mutex_lock(&write_mutex); 
        rt = do_readinode(e2fs, path, &ino, &inode);
	if (rt) {
		DEBUG_PRINTF("do_read indode %s failed\n", path);
		pthread_mutex_unlock(&write_mutex); 
                return rt;
	}

	/*
	 * Set file size
	 */
        efile->inode->i_size=inode.i_size;
        efile->inode->i_file_acl=inode.i_file_acl;
        pthread_mutex_unlock(&write_mutex); 
        

        fsize=inode.i_size;
        DEBUG_PRINTF("file_size %d %d %d GET %d\n",fsize,inode.i_size,efile->inode->i_size,ext2fs_file_get_size(efile)); 
         
        /*
	 * Truncate if required
	 */ 
        if(offset==0){ 
        if(inode.i_file_acl!=0)                                        
         {
          int length=offset; 
          char data[4096];
          memset(data,0,4096);
          int block=length/4096;
          short index=0;
          int *p; 
           
          if(length==0)
            {
                int acl_block;
                
                pthread_mutex_lock(&write_mutex); 
                io_channel_read_blk(e2fs->io,inode.i_file_acl,  //function for reading directly from physical block on disk
				    1, data);
                pthread_mutex_unlock(&write_mutex); 

                memcpy(&index,data,sizeof(short));
                memcpy(&acl_block,data+4092,sizeof(int)); 
               
                memset(data,0,4096);
                
                pthread_mutex_lock(&write_mutex); 
                io_channel_write_blk(e2fs->io,inode.i_file_acl, 1,data);
                pthread_mutex_unlock(&write_mutex);  

                printf("inode.i_file_acl truncated %d %d %d\n",acl_block,index,e2fs->block_map); 
                while(index==511){
                      int previous=acl_block;
                      printf("Truncating next block %d",acl_block);
                      
                      pthread_mutex_lock(&write_mutex);  
                      io_channel_read_blk(e2fs->io,acl_block,  //function for reading directly from physical block on disk
				    1, data);
                      pthread_mutex_unlock(&write_mutex);   

                      memcpy(&index,data,sizeof(short));
                      memcpy(&acl_block,data+4092,sizeof(int)); 
                      
                      pthread_mutex_lock(&write_mutex); 
                      ext2fs_block_alloc_stats(e2fs,previous, -1);  
                      pthread_mutex_unlock(&write_mutex); 

                    }

	       /*
                * For Debugging
                */	 	
               rt=ext2fs_test_block_bitmap (e2fs->block_map, 4812);
               DEBUG_PRINTF("TEST %d\n",rt);
               rt=ext2fs_test_block_bitmap (e2fs->block_map, 4809);
               DEBUG_PRINTF("TEST %d\n",rt);
               
               pthread_mutex_lock(&write_mutex);  
               rt=ext2fs_write_block_bitmap(e2fs);
               if(rt)
                 DEBUG_PRINTF("BITMAP WRITE FAILED\n");
               pthread_mutex_unlock(&write_mutex);  
                
             }
          else{               
                io_channel_read_blk(e2fs->io,inode.i_file_acl,1, data);

                blk_t acl_block=inode.i_file_acl;
                int *p,k;
                short items=0; 
                memcpy(&items,data,sizeof(short)); 
                for(index=0;index<items;)
                {
                 DEBUG_PRINTF("Truncate%d %d\n",index,items);
                 
                 p=(int *)(data+2+index*8);
                 if((*p)>block)
                   break;
                 

                 index++;
                 if(index==510 && items==511)
                      {
                            DEBUG_PRINTF("loading new ACL BLOCK %d\n",acl_block);
                            p=(int *)(data+4092); 
                            io_channel_read_blk(e2fs->io,*p,  //function for reading directly from physical block on disk
				    1, data);
                            acl_block=(blk_t)*p;
                            DEBUG_PRINTF("New ACl BLOCK %d\n",acl_block);  
                            index=0;
                                
                            memcpy(&items,data,sizeof(short));   
                      }
                  }                                                
 
             
             memcpy(data,&index,sizeof(short)); 
             io_channel_write_blk(e2fs->io,acl_block, 1,data);

             /*
              *Freeing up unused blocks
              */  
           
           while(items==511){
                 // free next block
                            DEBUG_PRINTF("Freeing blocks\n\n");
                            p=(int *)(data+4092); 
                            io_channel_read_blk(e2fs->io,*p,1, data);
                            memcpy(&items,data,sizeof(short));
                            ext2fs_block_alloc_stats(e2fs, *p, -1);  
                               

             }
           }
             DEBUG_PRINTF("index truncated to %d\n",index); 
          }

          }

	/*
	 * Set file size
	 */
         if (offset + size > fsize) {
                DEBUG_PRINTF("Set Size %d\n",offset+size); 
		
                pthread_mutex_lock(&write_mutex);  
                efile->inode->i_size = offset+size;
		efile->inode->i_size_high = 0;
                
		
                if (efile->ino) {
			DEBUG_PRINTF("Sttign %d %d %d\n\n",e2fs, efile->ino, efile->inode);
                        int retval = ext2fs_write_inode(e2fs, efile->ino, efile->inode);
			if (retval)
				{ 
				  DEBUG_PRINTF("SET SIZE failed\n\n");
                                  return retval;}
                         DEBUG_PRINTF("SET SIZE PASSED\n\n\n");
                        }
                 
                ext2fs_file_flush(efile); //flush to force writing of inode
                DEBUG_PRINTF("after size %d %d\n",ext2fs_file_get_size(efile),efile->inode->i_size); 
                pthread_mutex_unlock(&write_mutex);  
  
                rt = do_readinode(e2fs, path, &ino, &inode);
                DEBUG_PRINTF("after size %d ACL %d\n",inode.i_size,inode.i_file_acl); 
       
                
		if (rt) {
			DEBUG_PRINTF("do_readinode failed in do_write\n");
			return rt;
		}
	}

        pthread_mutex_lock(&write_mutex); 
	rt = ext2fs_file_llseek(efile, offset, 0, &npos);
	if (rt) {
		DEBUG_PRINTF("ext2fs_file_lseek(efile, %lld, SEEK_SET, &npos); failed\n\n", offset);
                pthread_mutex_unlock(&write_mutex); 
                return rt;
	}
        pthread_mutex_unlock(&write_mutex); 

	/*
	 * If encryption is turned off.
	 */
        if(e==0){

	for (rt = 0, wr = 0,wr1 = 0, tmp = buf; size > 0 && rt == 0; size -= wr, tmp += wr) {         //if encryption is off
		
                pthread_mutex_lock(&write_mutex); 
		rt = ext2fs_file_write(efile, tmp, size, &wr);
                pthread_mutex_unlock(&write_mutex);     

                DEBUG_PRINTF("WRITING  size: %u, written: %u", size, wr);
                if(!rt)
                    wr1+=wr; 
	   }
	}
	/*
	 * Encryption is turned on.
	 */
        else
        {
          struct region1{
           int block;
           short start;
           short end; 
          };  
         
          struct region1 abcd;

          char data[4096],*out;
          memset(data,0,4096);
          
          short index=0;
          int e_bytes=0; 
          int p1; 
          
	/*
	 * Allocate block to store encrypted data.
	 */
          if((inode.i_file_acl)==0){

           efile->inode->i_file_acl=inode.i_file_acl=alloc_block();
           
           pthread_mutex_lock(&write_mutex); 
           ext2fs_write_inode (e2fs,ino, &inode);
           pthread_mutex_unlock(&write_mutex); 

           DEBUG_PRINTF("ATTACHED ACL\n\n"); 
	   }
     

           DEBUG_PRINTF("Reading  ACL Block %d\n",inode.i_file_acl);
           pthread_mutex_lock(&write_mutex); 
           io_channel_read_blk(e2fs->io,inode.i_file_acl,1, data);
           pthread_mutex_unlock(&write_mutex); 
 

           memcpy(&index,data,sizeof(short));
           memcpy(&e_bytes,data+2+510*8,sizeof(int));

           DEBUG_PRINTF("OUTSIDE ATTACHED ACL\n\n");
           fflush(stdout); 

           int size1=0,block;
           int acl_block;

           acl_block=inode.i_file_acl;
 
           DEBUG_PRINTF("Before index %d\n", (int)index);           

           for (rt = 0, wr = 0, tmp = buf; size > 0 && rt == 0; size -= wr, tmp += wr)
           {
     
           block=offset/4096+1;        
           size1=size;
           if((block*4096-offset)<size)
             size1= block*4096-offset;

           out=malloc(size1);   
 	   /*
	    * Encrypt data.
	    */
           simple_encryption(tmp,1,size1,out);

           rt = ext2fs_file_write(efile, out, size1, &wr);

           DEBUG_PRINTF("WRITING size: %u, written: %u\n\n", size1, wr);

           if(!rt)
             wr1+=wr;

           DEBUG_PRINTF("After index %d\n", (int)index);           
 
           while(index>=510)
            {
                  DEBUG_PRINTF("index %d\n", (int)index);   
                  if(index==510){
		      /*
		       * Allocate a new block if the current block if full.
		       */
			
                      DEBUG_PRINTF("Writing index 511\n");
                      index=511;
                      memcpy(data,&index,sizeof(short));
                      p1=alloc_block();  
                      fflush(stdout);
                      DEBUG_PRINTF("Allocated block in write %d\n\n",p1);
                      memcpy(data+4092,&p1,sizeof(int));

                      pthread_mutex_lock(&write_mutex); 
                      io_channel_write_blk(e2fs->io,acl_block, 1,data);
                      io_channel_read_blk(e2fs->io,p1,1, data);
                      pthread_mutex_unlock(&write_mutex); 
                      e_bytes=0;
                      index=0;
                      acl_block=p1;
                      } 
                  else if(index==511){
		      /*
		       * Load next block.
		       */
                      DEBUG_PRINTF("Writing index 512\n");
                      int *z;
                      z=malloc(4);
                      memcpy(z,(data+4092),4);
                      DEBUG_PRINTF("Allocated block in write %d\n",*z); 
                      
                      pthread_mutex_lock(&write_mutex); 
                      io_channel_read_blk(e2fs->io,*z,1, data);
                      pthread_mutex_unlock(&write_mutex); 
 
                      memcpy(&index,data,sizeof(short));
                      memcpy(&e_bytes,data+2+510*8,sizeof(int));
                      int b=*z;
                      acl_block=b;
                      DEBUG_PRINTF("Allocated block in write %d %d INDEX %d\n",b,acl_block,index);
                      free(z);      
                   }
                
              }                           


              abcd.block=block;
              abcd.start=(offset%4096)+1;
              abcd.end=abcd.start+wr-1;
              e_bytes=e_bytes+abcd.end-abcd.start+1;
          

              index++;
          
              DEBUG_PRINTF("Writing abcd abcd %d %d %d INDEX %d BLOCK %d BYTES %d\n\n",abcd.block,abcd.start,abcd.end,index,acl_block,e_bytes); 

              memcpy(data,&index,sizeof(short));
              memcpy(data+(index-1)*sizeof(struct region1)+sizeof(short),&abcd,sizeof(struct region1));
              memcpy(data+510*sizeof(struct region1)+sizeof(short),&e_bytes,4);
            
              offset=offset+wr;
              free(out);
      
          }

        pthread_mutex_lock(&write_mutex); 
        io_channel_write_blk(e2fs->io,acl_block, 1,data);
        pthread_mutex_unlock(&write_mutex);         

       }
 


        

        fflush(stdout);  
	if (rt) {
		DEBUG_PRINTF("ext2fs_file_write(efile, tmp, size, &wr); failed");
		return -rt;
	}
        fflush(stdout);   
	pthread_mutex_lock(&write_mutex); 
        rt=ext2fs_file_flush(efile);
        pthread_mutex_unlock(&write_mutex); 
    
	if (rt) {
		DEBUG_PRINTF("ext2_file_flush(efile); failed %s..  ",path);
		return -rt;
	}
        
	return wr1;
}

